#lang racket

(provide (all-defined-out))

;; Zopakovanie.

(define (natural-number? x)
  (exact-nonnegative-integer? x))

;;; Univerzálna funkcia pre čiastočne rekurzívne funkcie.

;;; Aritmetizácia rekurzívnych termov a funkčných symbolov.

;; Konštruktory.

(define (Vt i) (list 'Vt i))               ; premenné
(define Zt (list 'Zt))                     ; konštanta 0
(define (Dt t1 t2 t3) (list 'Dt t1 t2 t3)) ; podmienený výraz
(define (Ct t ts) (list 'Ct t ts))         ; kontrakcia termov
(define (At e ts) (list 'At e ts))         ; aplikácia funkcie

(define Sx (list 'Sx))           ; funkcia nasledovníka
(define Px (list 'Px))           ; funkcia predchodcu
(define (Fx n) (list 'Fx n))     ; funkčná premenná
(define (Lx n t) (list 'Lx n t)) ; definovaná funkcia

;; Predikáty.

(define (Vt? t) (equal? (list-ref t 0) 'Vt))
(define (Zt? t) (equal? (list-ref t 0) 'Zt))
(define (Dt? t) (equal? (list-ref t 0) 'Dt))
(define (Ct? t) (equal? (list-ref t 0) 'Ct))
(define (At? t) (equal? (list-ref t 0) 'At))

(define (Sx? e) (equal? (list-ref e 0) 'Sx))
(define (Px? e) (equal? (list-ref e 0) 'Px))
(define (Fx? e) (equal? (list-ref e 0) 'Fx))
(define (Lx? e) (equal? (list-ref e 0) 'Lx))

;; Selektory.

(define (Vt-index t) (list-ref t 1)) ; index premennej
(define (Dt-if t) (list-ref t 1))    ; podmienka
(define (Dt-then t) (list-ref t 2))  ; dôsledok
(define (Dt-else t) (list-ref t 3))  ; alternatíva
(define (Ct-first t) (list-ref t 1)) ; prvý term
(define (Ct-rest t) (list-ref t 2))  ; zvyšné termy
(define (At-fun t) (list-ref t 1))   ; funkcia
(define (At-arg t) (list-ref t 2))   ; argument(y)

(define (Fx-arity e) (list-ref e 1)) ; arita funkčnej premennej
(define (Lx-arity e) (list-ref e 1)) ; arita definovanej funkcie
(define (Lx-body e) (list-ref e 2))  ; pravá strana definičnej rovnosti

;; Údajový typ pre rekurzívne termy a funkčné symboly.

(define (TERM? t)
  (and (list? t)
       (>= (length t) 1)
       (or (and (Vt? t)
                (= (length t) 2)
                (natural-number? (Vt-index t)))
           (and (Zt? t)
                (= (length t) 1))
           (and (Dt? t)
                (= (length t) 4)
                (TERM? (Dt-if t))
                (TERM? (Dt-then t))
                (TERM? (Dt-else t)))
           (and (Ct? t)
                (= (length t) 3)
                (TERM? (Ct-first t))
                (TERM? (Ct-rest t)))
           (and (At? t)
                (= (length t) 3)
                (FUNCTION? (At-fun t))
                (TERM? (At-arg t))))))

(define (FUNCTION? e)
  (and (list? e)
       (>= (length e) 1)
       (or (and (Sx? e)
                (= (length e) 1))
           (and (Px? e)
                (= (length e) 1))
           (and (Fx? e)
                (= (length e) 2)
                (natural-number? (Fx-arity e)))
           (and (Lx? e)
                (= (length e) 3)
                (natural-number? (Lx-arity e))
                (TERM? (Lx-body e))))))

;; Predikát recursive-function-symbol?(n,e) platí, ak číslo e je kódom n-árneho
;; rekurzívneho funkčného symbolu.

(define (recursive-function-symbol? n e)
  (define (is-recursive-function-symbol? n e)
    (or (and (Sx? e) (= n 1))
        (and (Px? e) (= n 1))
        (and (Lx? e)
             (= (Lx-arity e) n)
             (recursive-term? n (Lx-body e)))))
  (and (FUNCTION? e) (>= n 1) (is-recursive-function-symbol? n e)))

(define (recursive-term? n t)
  (define (is-recursive-term? n t)
    (or (and (Vt? t)
             (<= 1 (Vt-index t))
             (<= (Vt-index t) n))
        (Zt? t)
        (and (Dt? t)
             (is-recursive-term? n (Dt-if t))
             (is-recursive-term? n (Dt-then t))
             (is-recursive-term? n (Dt-else t)))
        (and (At? t)
             (let* ([e (At-fun t)] [m (arity e)])
               (if (Fx? e)
                   (and (= m n)
                        (are-recursive-terms? n (At-arg t) n))
                   (and (>= m 1)
                        (recursive-function-symbol? m e)
                        (are-recursive-terms? n (At-arg t) m)))))))
  (define (are-recursive-terms? n ts k)
    (or (and (= k 1)
             (is-recursive-term? n ts))
        (and (> k 1)
             (Ct? ts)
             (is-recursive-term? n (Ct-first ts))
             (are-recursive-terms? n (Ct-rest ts) (- k 1)))))
  (and (TERM? t) (is-recursive-term? n t)))

;; Arita funkčných symbolov.

(define (arity e)
  (cond [(Sx? e) 1]
        [(Px? e) 1]
        [(Fx? e) (Fx-arity e)]
        [(Lx? e) (Lx-arity e)]))

;;; Aritmetizácia redukcií.

;; Program, ktorý pre dané číslo vytvorí odpovedajúci numerál.

(define (number->numeral x)
  (if (= x 0) Zt (At Sx (number->numeral (- x 1)))))
